#include "dominodevice.h"

#include <stdexcept>
#include <QtDebug>
#include <QString>
#include <QMessageBox>
#include <QDesktopServices>
#include <QDir>

#include "dominodata.h"
#include "drs.h"
#include "regio/mlinkdevice.h"

const int MAX_MEMRD_SIZE = 256;       //in 32bit words
const int ADC_BITS = 12;
const int ADC_HALF_VALUE = 1<<(ADC_BITS-1);
const int ADC_MASK = (1<<ADC_BITS) - 1;
const int DRS_FIRST_CELL = 13;

using mlink::RegOpVector;
using mlink::RegWrite;

enum {
    AD9252_SPI_12BITS_MODE = 0x3
};

enum {
    AD9252_SPI_DEVICE_UPDATE_BIT = 0x1
};

enum {
    AD9252_SPI_RESET = 0x3C,
    AD9252_SPI_RUN = 0x18
};


enum {
    AD9252_SPI_CHIP_PORT_CONFIG_REG = 0x0,
    AD9252_SPI_SERIAL_CONTROL_REG = 0x21,
    AD9252_SPI_DEVICE_UPDATE_REG = 0xFF
};

enum {
    // 1
    REG_SWITCH_CONTROL = 0x108, // 8
    REG_SUPERDAC_TEST = 0x109, // 9
    REG_SUPERDAC_OUTPUT_DATA = 0x10A, // 10
    // 13
    // 14
    // 15
    REG_DAC1 = 0x110, // 16
    REG_DAC2 = 0x111, // 17

    REG_SUPERDAC_SPI_ADDR = 0x118,// 24
    REG_SUPERDAC_SPI_DATA1 = 0x119,// 25
    REG_SUPERDAC_SPI_DATA2 = 0x11A,// 26

    REG_AD9252_SPI_ADDR = 0x120, // 288
    REG_AD9252_SPI_WR_DATA = 0x121, // 289
    REG_AD9252_SPI_READ_DATA = 0x122, // 290

    //Domino regs
    REG_0 = 0x400,  //run_ctrl
    REG_3 = 0x403,
    REG_4 = 0x404,
    REG_DELAY_LENGTH = 0x405, // 5
    REG_RD_MAX_CNT = 0x406, // 6
    REG_WR_MAX_CNT = 0x407, // 7
    REG_CHANNEL_CAN_BE_TRIGGER = 0x40B, //11
    REG_TRIGGER_POLARITY = 0x40C, // 12
    REG_DRS_CAN_TRIG = 0x412, // 18
    REG_DRS_AFTER_TRIG = 0x413, // 19
    REG_DRS_PAUSE_BEFORE_READ = 0x414, // 20
    REG_DRS_READ_COUNTER = 0x415, // 21
    REG_DRS_PAUSE_BEFORE_MEM_WRITE = 0x416 // 22
};

DominoDevice::DominoDevice(QObject *parent)
    : QtMregDevice(parent)
{
    addSupportedDeviceId(QtMregDevice::DEVICE_ID_TEST);
    addSupportedDeviceId(QtMregDevice::DEVICE_ID_EVADC);
    addSupportedDeviceId(QtMregDevice::DEVICE_ID_ADC8BEDRS);
}

//void DominoDevice::setSetup(const DominoSettings &r)
//{
//    setup = r;
//}

void DominoDevice::writeSettings(const DominoSettings &in, bool init)
{
    setup = in;
    if (setup.mode == DominoSettings::AMPL_CAL_MODE) {
        setup.writecellnumber = DRS_SIZE;
        setup.readcellnumber = DRS_SIZE;
        for (int i=0; i<DRS_CHANNELS; i++) {
//            setup.thr[i] = 0;
            setup.en[i] = true;
            setup.trig[i] = true;
        }
    }
    RegOpVector v;
    for (int i=0; i<DRS_CHANNELS; i++) {
        v += DacWrite(1, DRS_CHANNELS-i, setup.thr[i]);
    }

    v.push_back(RegWrite(REG_CHANNEL_CAN_BE_TRIGGER, setup.getTrigMask()));
    v.push_back(RegWrite(REG_WR_MAX_CNT, setup.writecellnumber));
    v.push_back(RegWrite(REG_RD_MAX_CNT, setup.readcellnumber));
    v.push_back(RegWrite(REG_DRS_AFTER_TRIG, setup.trigdelay));
    v.push_back(RegWrite(REG_DELAY_LENGTH, setup.analogDelay));

    if (init)
    {
        v += SetMode();
        v += deviceInit();
        ad9252_setup(); // All setup implemented in hardware
    }
    regOpExec(v);
}

RegOpVector DominoDevice::deviceInit()
{
    RegOpVector v;

    v += DacWrite(2, 2, 1590);  //set OOFS
    v += DacWrite(2, 1, 1980);  //set ROFS for evadc
//    v += DacWrite(2, 1, 1770);  //set ROFS for adc8be-drs
    v += DacWrite(2, 4, 1680);  //set IAMP
    v += DacWrite(2, 6, 1680);  //set SDAC IAMP
    v += DacWrite(2, 5, 1680);  //same for compatibility with ADC8BE-DRS

    v.push_back(RegWrite(REG_DRS_READ_COUNTER, DRS_SIZE+1));     //number of drs analog data to read
    return v;
}

RegOpVector DominoDevice::SetMode()
{
    RegOpVector v;
    int mode = setup.mode;
    switch(mode)
    {
    case DominoSettings::NORMAL_MODE:
    {
        v.push_back(RegWrite(REG_SWITCH_CONTROL, 0)); //analog MUX to real signal
        v.push_back(RegWrite(REG_SUPERDAC_TEST, 0)); //Disable SDAC;
        break;
    }
    case DominoSettings::AMPL_CAL_MODE:
    {
        v.push_back(RegWrite(REG_SWITCH_CONTROL, 0x1)); //analog MUX to SDAC
        v.push_back(RegWrite(REG_SUPERDAC_TEST, 0x1)); //Enable SDAC;
        v.push_back(RegWrite(REG_SUPERDAC_OUTPUT_DATA, 0x1D00)); //Value for SDAC
        break;
    }
    }
    return v;
}

DominoData DominoDevice::applyAmpCalibration(const DominoData &in) const
{
    DominoData dominoData(in);
    Q_FOREACH(int ch, in.channelData.keys()) {
                //qDebug() << "apply calibration to channel" << ch;
        const QVector<double> &Aclb = cellCalibration.getCellAmpCalibration(ch);
        Q_ASSERT(Aclb.size() == DRS_SIZE);
        for (int i=0; i<dominoData.channelData[ch].Data.size(); i++)
        {
            int realcell = (dominoData.StartCell + i) % DRS_SIZE;
            if (i==0) realcell = (realcell+ 1) % DRS_SIZE; //taking calibration from next cell for zero point (data[0] == data[1] see in ReadChMem )
            dominoData.channelData[ch].Data[i] -= Aclb[realcell];
        }
    }
    return dominoData;
}

void DominoDevice::CalibrateCellAmp(const DominoData &dominoData)
{
    Q_FOREACH(int ch, dominoData.channelData.keys()) {
        //        if(!ChEn[ch]) continue;
//        qDebug() << "calibrating channel " << ch;
        cellCalibration.addCellAmpData(ch, dominoData.StartCell, dominoData.channelData[ch].rawData);
    }
}

void DominoDevice::StartRun()
{
    regWrite(REG_0, setup.getEnMask());
}

void DominoDevice::SaveData(const DominoData &dominoData)
{
    DataOut << "Event timestamp " << QString("%1").arg(comp_time.currentDateTime().toTime_t()) <<
               " DRS first cell " << dominoData.StartCell << "\n";
    Q_FOREACH(int i, dominoData.channelData.keys())
    {
        if(!setup.en[i]) continue;
        const DominoChannelData &channelData = dominoData.channelData[i];
        DataOut << i+1 << " ";
        for (int j=0; j<channelData.Data.size(); j++)
            DataOut << QString("%1").arg(channelData.Data[j], 0, 'e') << " ";
        DataOut << "\n";
    }
}

void DominoDevice::StopRun()
{
    regWrite(REG_0, 0);
}

DominoChannelData DominoDevice::ReadChMem(int ch)
{
    DominoChannelData channelData;
    //    Q_ASSERT(ch < dominoData.Data.size());
    const size_t evenReadWordsNum = (setup.readcellnumber+1)/2 * 2;
    const size_t readSize = evenReadWordsNum/2;

    quint16 readaddr = 0;

    const std::vector<quint32> &rd = memReadBlk((ch << 11) | readaddr, readSize);
    if (readSize != rd.size())
    {
        qCritical() << "read memory size mismatch";
        return channelData;
    }
    std::vector<qint16> unrolled;
    unrolled.reserve(rd.size() * 2);
    for (size_t i=0; i < rd.size(); i++)
    {
        unrolled.push_back((int(rd[i]) & ADC_MASK) - ADC_HALF_VALUE);
        unrolled.push_back((int(rd[i]>>16) & ADC_MASK) - ADC_HALF_VALUE);        
    }
    unrolled[0] = unrolled[1]; // 1st point in DRS data sometimes wrong(real bad data in drs cell, checked with twice reading domino ring)
    Q_ASSERT(evenReadWordsNum == unrolled.size());
    channelData.rawData.resize(unrolled.size()/*evenReadWordsNum*/);
    for(size_t i=0; i<unrolled.size(); i++)
        channelData.rawData[i] = unrolled[i];
//        channelData.rawData[(unsigned int)(i+startCell) % DRS_SIZE] = unrolled[i];
//    qDebug() << "rawData size" << channelData.rawData.size();

    for (int i=0; i< channelData.rawData.size(); i++) {
//        channelData.Data.push_back(channelData.rawData[i*2]);
//        channelData.Data.push_back(channelData.rawData[i*2+1]);
        channelData.Data.push_back(channelData.rawData[i]);
    }
    //    channelData.rawData.resize(ReadWordsNum);
    //    channelData.Data.resize(ReadWordsNum);
//    qDebug() << "Data size" << channelData.Data.size();
    return channelData;
}

DominoData DominoDevice::ReadAllChMem()
{
    DominoData dominoData;
    //    dominoData.Data.resize(domino_channels);
    //    dominoData.rawData.resize(domino_channels);
    dominoData.StartCell = regRead(REG_3) & DRS_CELL_NUMBER_MASK;
    //qDebug() << "Starting cell number is " << dominoData.StartCell;
    for(int ch=0; ch<DRS_CHANNELS; ch++)
        if (setup.en[ch])
            dominoData.channelData[ch] = ReadChMem(ch);
    dominoData.EventNumber++;
    return dominoData;
}

bool DominoDevice::isRunning()
{
//    qDebug() << "still running" << " reg0=" << regRead(REG_0);
    readTemp();
    return 0 != regRead(REG_0);
}

RegOpVector DominoDevice::DacWrite(int dacnum, int ch, int data)
{
    Q_ASSERT(dacnum >= 1 && dacnum <= 2);
    RegOpVector v;
    v.push_back(RegWrite(REG_DAC1 + dacnum - 1, ((((ch-1)&0x7)<<12) | (data&0xFFF))));
    return v;
}

void DominoDevice::ad9252_spi_write(quint16 addr, quint8 data)
{
    regWrite(REG_AD9252_SPI_WR_DATA, data);
    regWrite(REG_AD9252_SPI_ADDR, addr & 0x1FFF);   //bits 13,14,15 = 0 (write and W0=W1=0 - 1 byte operation)
}

quint16 DominoDevice::ad9252_spi_read(quint16 addr)
{
    regWrite(REG_AD9252_SPI_ADDR, ((addr&0x9FFF) | 0x8000));   //bits 13,14 = 0 , bit15 = 1 (read and W0=W1=0 - 1 byte operation)
    return regRead(REG_AD9252_SPI_READ_DATA);
}

void DominoDevice::ad9252_setup()
{
// All setup implemented in hardware
/*
      ad9252_spi_write(AD9252_SPI_CHIP_PORT_CONFIG_REG, AD9252_SPI_RESET);
      const int reset_watchdog = 10;
      int watchdog = 0;
      while(true)
      {
          if (ad9252_spi_read(AD9252_SPI_CHIP_PORT_CONFIG_REG) == AD9252_SPI_RUN) break;
          watchdog++;
          if(watchdog > reset_watchdog)
          {
              qCritical() << "AD9252 reset failed";
              break;
          }
      }
    if (ad9252_spi_read(AD9252_SPI_SERIAL_CONTROL_REG) != AD9252_SPI_12BITS_MODE)
    {
        ad9252_spi_write(AD9252_SPI_SERIAL_CONTROL_REG, AD9252_SPI_12BITS_MODE);
        ad9252_spi_write(AD9252_SPI_DEVICE_UPDATE_REG, AD9252_SPI_DEVICE_UPDATE_BIT);
    }
*/
}

void DominoDevice::readCellCalibFromFile()
{
    QVector<QVector<double> > Aclb;
    Aclb.resize(DRS_CHANNELS);
    for(int i=0; i<DRS_CHANNELS; i++)
        Aclb[i].resize(DRS_SIZE);  // 0 value will be used, if something is wrong in file

    QFile data(QDesktopServices::storageLocation(QDesktopServices::DataLocation) + "/cellamp.clb");
    if (data.open(QFile::ReadOnly)) {
        QTextStream in(&data);
        for(int i=0; i<DRS_SIZE; i++)
        {
            for(int j=0; j<DRS_CHANNELS; j++)
                in >> Aclb[j][i];
        }
    }
    else
    {
        qDebug() << "cell amp. calibration read from file failed";
        return;
    }
    data.close();
    for(int i=0; i<DRS_CHANNELS; i++)
        cellCalibration.addCellAmpCalibration(i,Aclb[i]);
}

void DominoDevice::writeCellCalibToFile()
{
    QDir mydir;
    QVector<QVector<double> > Aclb;
    Aclb.resize(DRS_CHANNELS);

    QString filename = QDesktopServices::storageLocation(QDesktopServices::DataLocation);
    if(!mydir.exists(filename)) mydir.mkpath(filename);
    filename += "/cellamp.clb";
    QFile data(filename);
    if (data.open(QFile::WriteOnly | QFile::Truncate)) {
        QTextStream out(&data);
        for(int i=0; i<DRS_CHANNELS; i++)
            Aclb[i] = cellCalibration.getCellAmpCalibration(i);
        for(int i=0; i<DRS_SIZE; i++)
        {
            for(int j=0; j<DRS_CHANNELS; j++)
                out << Aclb[j][i] << " ";
            out << "\n";
        }
    }
    else
    {
        qDebug() << "cell amp. calibration write to file failed";
        return;
    }
    data.close();
}

bool DominoDevice::RegWRTest(int NCycle)
{
    qDebug() << "My test point..." << NCycle;
}
