#include "mainwindow.h"

#include "ui_mainwindow.h"
#include "dominodevice.h"
#include <regio/regiomlink.h>

#include <QtDebug>
#include <QTimer>
#include <QHostAddress>
#include <qxw/WaveformPlot.h>
#include <stdexcept>
#include <QFileDialog>
#include <QMessageBox>
#include <QSettings>
#include <QStatusBar>
#include <QDesktopServices>
#include <QUrl>
#include <QVector>
#include "drs.h"
#include "util/QDateTimeCompat.h"

#include "waitcursor.h"

const int displayUpdateTimeMs = 30;

MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow),
    dominoDevice(new DominoDevice(this))
{
    ui->setupUi(this);
    ui->spinBoxReadCNT->setMaximum(DRS_SIZE);
    ui->spinBoxWriteCNT->setMaximum(DRS_SIZE);
    statusLabel = new QLabel(this);
    statusBar()->addPermanentWidget(statusLabel);
    connect(dominoDevice, SIGNAL(deviceConnected()), SLOT(onDeviceConnect()));
    connect(dominoDevice, SIGNAL(setStatus(const QString &)), SLOT(setStatus(const QString &)));
    connect(dominoDevice, SIGNAL(TempChanged(double)), SLOT(setTemp(double)));
    readSettings();
    readCellAmpCalib();
    restoreFormValues();
    tmr = new QTimer(this);
    tmr->setSingleShot(true);
    connect(tmr, SIGNAL(timeout()), SLOT(DoTimer()));
    ui->groupBoxWRReg->hide();
    ui->groupBoxDAC->hide();
////    ui->pushButtonTest2->hide();
    ui->xPlot->setAxisTitle(QwtPlot::xBottom, "Time, ns");
    ui->xPlot->setAxisScale(QwtPlot::xBottom, 0., DRS_TIME_BIN * 1e9 * DRS_SIZE);
    qRegisterMetaType<DominoRealData>();
    //    connect(this, SIGNAL(gotData(DominoData)), SLOT(appendCalibrationData(DominoData)));
    connect(this, SIGNAL(gotData(DominoData)), SLOT(DrawScope(DominoData)));
    connect(this, SIGNAL(gotData(DominoData)), SLOT(saveData(DominoData)));
}

MainWindow::~MainWindow()
{
    delete dominoDevice;
    delete ui;
}

void MainWindow::changeEvent(QEvent *e)
{
    QMainWindow::changeEvent(e);
    switch (e->type()) {
    case QEvent::LanguageChange:
        ui->retranslateUi(this);
        break;
    default:
        break;
    }
}

void MainWindow::closeEvent(QCloseEvent *event)
{
    writeSettings();
    QMainWindow::closeEvent(event);
}

void MainWindow::on_actionQuit_activated()
{
    close();
}

void MainWindow::on_actionDocs_activated()
{
    QDesktopServices::openUrl(QUrl("http://afi.jinr.ru/EVADC"));
}

void MainWindow::on_actionCheckForUpdates_activated()
{
    QDesktopServices::openUrl(QUrl("http://afi.jinr.ru/Downloads/EVADC"));
}

void MainWindow::on_actionAbout_activated()
{
    QString rev;
#ifdef SVN_REV
    rev = QString("%1").arg(SVN_REV);
    rev = rev.remove(QRegExp("^.*\\:"));
#endif
    if (!rev.isEmpty())
        rev = QString(", rev. %1").arg(rev);
    QString s = "<p>EVADC<br>Version 0.1"+rev+"<br><br>&copy; 2010 <a href=\"http://afi.jinr.ru/\">AFI Electronics</a><br>JINR, Dubna";
    QMessageBox::about (this, "About", s);
}

void MainWindow::on_actionAboutQt_activated()
{
    qApp->aboutQt();
}

void MainWindow::setStatus(const QString &s)
{
    statusLabel->setText(s);
}

void MainWindow::onDeviceConnect()
{
    lastSuccessfulHostAddress = ui->lineEditHostName->text();
}

void MainWindow::onDeviceDisconnect()
{
    QMessageBox::critical(0, "Error", "Device disconnected");
}

void MainWindow::setTemp(double value)
{
    ui->labelTemp->setText(QString("t%1C").arg(value));
}

void MainWindow::DoStopTimer()
{
    //    qDebug() << "stop timer";
    LocalWaitCursor wc(this);
    dominoDevice->StopRun();
    ui->pushButtonStart->setEnabled(true);
    ui->checkBoxW2D->setEnabled(true);
}

void MainWindow::DoTimer()
{
    try
    {
//        LocalWaitCursor wc(this);
        dominoDevice->writeSettings(setup, false);  //only refresh thresholds, triggers etc. (don't update domino settings)
        if(!dominoDevice->isRunning())
        {
            const DominoData &dominoData = dominoDevice->ReadAllChMem();
            appendCalibrationData(dominoData);
            emit gotData(dominoData);
            if(ui->checkBoxSnapshot->isChecked()) on_pushButtonStop_clicked();
            else dominoDevice->StartRun();
        }
        if (dominoDevice->ActiveRun) tmr->start(1);
    }catch(std::runtime_error& e){ qDebug() << e.what(); }

}

void MainWindow::DrawScope(const DominoData &in)
{
    if (QDateTimeMsescTo(scopeUpdateTime, QDateTime::currentDateTime()) < displayUpdateTimeMs) return;
    const DominoData &dominoData = applyAmpCalibration(in);
//    const DominoData &dominoData = in;
    WaveformDataAll mvv;
    Q_FOREACH(int ch, dominoData.channelData.keys())
    {
        if (!setup.en[ch])
            continue;
        const DominoChannelData &channelData = dominoData.channelData[ch];
        const DominoRealData &realData = channelData.Data;
        const QString name = QString("ch%1").arg(ch);
        std::vector<double> vx;
        for(int i=0; i<realData.size(); i++)
            vx.push_back(i * DRS_TIME_BIN * 1e9);
//            vx.push_back(i);
        WaveformData wd(vx, realData.toStdVector(), name.toStdString(), ch);
        mvv.push_back(wd);
    }
    ui->xPlot->setAllData(mvv);
    scopeUpdateTime = QDateTime::currentDateTime();
}

void MainWindow::appendCalibrationData(const DominoData &dominoData)
{
    if (ui->radioButtonAOffMod->isChecked())
        dominoDevice->CalibrateCellAmp(dominoData);
}

DominoData MainWindow::applyAmpCalibration(const DominoData &dominoData) const
{
    return dominoDevice->applyAmpCalibration(dominoData);
}

void MainWindow::saveData(const DominoData &dominoData)
{
    if (ui->checkBoxW2D->isChecked()) dominoDevice->SaveData(applyAmpCalibration(dominoData));
}

void MainWindow::on_pushButtonStart_clicked()
{
    LocalWaitCursor wc(this);
    if (setup.getEnMask() == 0)
    {
        QMessageBox::warning(this, tr("EVADC Warning"), tr("No channel selected"), QMessageBox::Ok);
        return;
    }
    dominoDevice->writeSettings(setup, true);  //full setup (inc. domino)
    dominoDevice->StartRun();
    dominoDevice->ActiveRun = true;
    ui->pushButtonStart->setEnabled(false);
    ui->checkBoxW2D->setEnabled(false);
    tmr->start(1);
}

void MainWindow::on_pushButtonStop_clicked()
{
    dominoDevice->ActiveRun = false;
    tmr->stop();
    QTimer::singleShot(300, this, SLOT(DoStopTimer()));
}

void MainWindow::on_pushButton_t1_clicked()
{
    dominoDevice->regWrite( ui->spinBox_reg->value() , ui->spinBox_data->value() );
}

void MainWindow::on_pushButton_t2_clicked()
{
    qDebug() << "reg " << ui->spinBox_reg->value() << "=" << hex << dominoDevice->regRead( ui->spinBox_reg->value());
}

void MainWindow::on_pushButton_rm_clicked()
{
    LocalWaitCursor wc(this);
    dominoDevice->RegWRTest(1);
    //    while(true) {
/*
    std::vector<quint32> wr;
    for(int i=0; i<5; i++)
        wr.push_back(rand());
    qDebug() << "Writing RAM";
    for(size_t i=0; i<wr.size(); i++)
        qDebug() << QString("wr[%1] = %2").arg(i).arg(wr[i]);
    dominoDevice->memWriteBlk(0x0, wr);
    qDebug() << "Reading RAM";
    const QVector<quint32> &rd = QVector<quint32>::fromStdVector(dominoDevice->memReadBlk(0x0, wr.size()));
    for(int i=0; i<rd.size(); i++)
        qDebug() << QString("rd[%1] = %2").arg(i).arg(rd[i]);
    //    dominoDevice->ReadAllChMem();
    //    DrawScope();
    qDebug() << "end of memory read";
    //    }
*/
/*
    qDebug() << "Mem Test started";
    QVector<quint32> rd;
    for (int i=0; i<500000; i++)
        rd = QVector<quint32>::fromStdVector(dominoDevice->memReadBlk(0x0, 340));
    qDebug() << "Mem Test finished";
*/
}

void MainWindow::on_pushButtonTest2_clicked()
{
    dominoDevice->writeCellCalibToFile();
}

void MainWindow::on_pushButton_setdac_clicked()
{
//    LocalWaitCursor wc(this);
//    dominoDevice->DacWrite(ui->spinBox_dacnum->value(), ui->spinBox_dacchannelnum->value(), ui->spinBox_dacdata->value());
}

void MainWindow::on_checkBoxW2D_toggled(bool checked)
{
    if(checked)
    {
        dominoDevice->DataFileName = QFileDialog::getSaveFileName(this,
                                                                  tr("Save File"),
                                                                  QDir::homePath()+QDir::separator()+"evadc8.txt",
                                                                  tr("Text files *.txt;;All files *.*"));
        dominoDevice->DataFile.setFileName(dominoDevice->DataFileName);
        if (dominoDevice->DataFile.open(QFile::WriteOnly | QFile::Truncate))
        {
            //dominoDevice->DataOut.setDevice();
            dominoDevice->DataOut.setDevice(&dominoDevice->DataFile);

        }
        else ui->checkBoxW2D->setChecked(false);
    }
    else dominoDevice->DataFile.close();
}

void MainWindow::restoreFormValues()
{
    ui->checkBoxE1->setChecked(setup.en[0]);
    ui->checkBoxE2->setChecked(setup.en[1]);
    ui->checkBoxE3->setChecked(setup.en[2]);
    ui->checkBoxE4->setChecked(setup.en[3]);
    ui->checkBoxE5->setChecked(setup.en[4]);
    ui->checkBoxE6->setChecked(setup.en[5]);
    ui->checkBoxE7->setChecked(setup.en[6]);
    ui->checkBoxE8->setChecked(setup.en[7]);

    ui->spinBoxT1->setValue(setup.thr[0]);
    ui->spinBoxT2->setValue(setup.thr[1]);
    ui->spinBoxT3->setValue(setup.thr[2]);
    ui->spinBoxT4->setValue(setup.thr[3]);
    ui->spinBoxT5->setValue(setup.thr[4]);
    ui->spinBoxT6->setValue(setup.thr[5]);
    ui->spinBoxT7->setValue(setup.thr[6]);
    ui->spinBoxT8->setValue(setup.thr[7]);

    ui->checkBoxTr1->setChecked(setup.trig[0]);
    ui->checkBoxTr2->setChecked(setup.trig[1]);
    ui->checkBoxTr3->setChecked(setup.trig[2]);
    ui->checkBoxTr4->setChecked(setup.trig[3]);
    ui->checkBoxTr5->setChecked(setup.trig[4]);
    ui->checkBoxTr6->setChecked(setup.trig[5]);
    ui->checkBoxTr7->setChecked(setup.trig[6]);
    ui->checkBoxTr8->setChecked(setup.trig[7]);

    ui->radioButtonNormMode->setChecked(setup.mode == DominoSettings::NORMAL_MODE);
    ui->radioButtonAOffMod->setChecked(setup.mode == DominoSettings::AMPL_CAL_MODE);

    ui->spinBoxWriteCNT->setValue(setup.writecellnumber);
    ui->spinBoxReadCNT->setValue(setup.readcellnumber);
    ui->spinBoxAfterTrig->setValue(setup.trigdelay);
    ui->spinBoxADelay->setValue(setup.analogDelay);
}

void MainWindow::on_checkBoxE1_toggled(bool checked)
{ setup.en[0] = checked; }
void MainWindow::on_checkBoxE2_toggled(bool checked)
{ setup.en[1] = checked; }
void MainWindow::on_checkBoxE3_toggled(bool checked)
{ setup.en[2] = checked; }
void MainWindow::on_checkBoxE4_toggled(bool checked)
{ setup.en[3] = checked; }
void MainWindow::on_checkBoxE5_toggled(bool checked)
{ setup.en[4] = checked; }
void MainWindow::on_checkBoxE6_toggled(bool checked)
{ setup.en[5] = checked; }
void MainWindow::on_checkBoxE7_toggled(bool checked)
{ setup.en[6] = checked; }
void MainWindow::on_checkBoxE8_toggled(bool checked)
{ setup.en[7] = checked; }

void MainWindow::on_spinBoxT1_valueChanged(int newval)
{ setup.thr[0] = newval; }
void MainWindow::on_spinBoxT2_valueChanged(int newval)
{ setup.thr[1] = newval; }
void MainWindow::on_spinBoxT3_valueChanged(int newval)
{ setup.thr[2] = newval; }
void MainWindow::on_spinBoxT4_valueChanged(int newval)
{ setup.thr[3] = newval; }
void MainWindow::on_spinBoxT5_valueChanged(int newval)
{ setup.thr[4] = newval; }
void MainWindow::on_spinBoxT6_valueChanged(int newval)
{ setup.thr[5] = newval; }
void MainWindow::on_spinBoxT7_valueChanged(int newval)
{ setup.thr[6] = newval; }
void MainWindow::on_spinBoxT8_valueChanged(int newval)
{ setup.thr[7] = newval; }

void MainWindow::on_checkBoxTr1_toggled(bool checked)
{ setup.trig[0] = checked; }
void MainWindow::on_checkBoxTr2_toggled(bool checked)
{ setup.trig[1] = checked; }
void MainWindow::on_checkBoxTr3_toggled(bool checked)
{ setup.trig[2] = checked; }
void MainWindow::on_checkBoxTr4_toggled(bool checked)
{ setup.trig[3] = checked; }
void MainWindow::on_checkBoxTr5_toggled(bool checked)
{ setup.trig[4] = checked; }
void MainWindow::on_checkBoxTr6_toggled(bool checked)
{ setup.trig[5] = checked; }
void MainWindow::on_checkBoxTr7_toggled(bool checked)
{ setup.trig[6] = checked; }
void MainWindow::on_checkBoxTr8_toggled(bool checked)
{ setup.trig[7] = checked; }

void MainWindow::on_radioButtonNormMode_toggled(bool checked)
{
    if (checked) {
        setup.mode = DominoSettings::NORMAL_MODE;
    }
}

void MainWindow::on_radioButtonAOffMod_toggled(bool checked)
{
    if (checked) {
        setup.mode = DominoSettings::AMPL_CAL_MODE;
        dominoDevice->clearCellCalib();
    }
}

void MainWindow::on_spinBoxReadCNT_valueChanged(int newval)
{ setup.readcellnumber=newval; }

void MainWindow::on_spinBoxWriteCNT_valueChanged(int newval)
{ setup.writecellnumber=newval; }

void MainWindow::on_spinBoxAfterTrig_valueChanged(int newval)
{ setup.trigdelay=newval; }

void MainWindow::on_spinBoxADelay_valueChanged(int newval)
{
    setup.analogDelay = newval;
}

void MainWindow::readCellAmpCalib()
{
    dominoDevice->readCellCalibFromFile();
}

void MainWindow::readSettings()
{
    QSettings settings;
    settings.beginGroup("MainWindow");
    restoreGeometry(settings.value("geometry").toByteArray());
    restoreState(settings.value("windowState").toByteArray());
    QString s = settings.value("boardHostAddress").toString();
    lastSuccessfulHostAddress = s;
    if (!s.isNull()) {
        ui->lineEditHostName->setText(s);
        setBoardHostName();
    }
    for(int i=0; i<DRS_CHANNELS; i++) {
        setup.en[i] = settings.value(QString("setup.en[%1]").arg(i), true).toBool();
        setup.thr[i] = settings.value(QString("setup.thr[%1]").arg(i),800).toInt();
        setup.trig[i] = settings.value(QString("setup.trig[%1]").arg(i), true).toBool();
    }
    setup.mode = settings.value("mode", DominoSettings::NORMAL_MODE).toInt();
    setup.readcellnumber = settings.value("readcellnumber", DRS_SIZE).toInt();
    setup.writecellnumber = settings.value("writecellnumber", DRS_SIZE).toInt();
    setup.trigdelay = settings.value("trigdelay",5).toInt();
    settings.endGroup();
}

void MainWindow::resetSettings()
{
    for(int i=0; i<DRS_CHANNELS; i++) {
        setup.en[i] = true;
        setup.thr[i] = 800;
        setup.trig[i] = true;
    }
    setup.mode = DominoSettings::NORMAL_MODE;
    setup.readcellnumber = DRS_SIZE;
    setup.writecellnumber = DRS_SIZE;
    setup.trigdelay = 5;
}

void MainWindow::writeSettings()
{
    QSettings settings;
    settings.beginGroup("MainWindow");
    settings.setValue("geometry", saveGeometry());
    settings.setValue("windowState", saveState());
    settings.setValue("boardHostAddress", lastSuccessfulHostAddress);
    for(int i=0; i<DRS_CHANNELS; i++) {
        settings.setValue(QString("setup.en[%1]").arg(i), bool(setup.en[i]));
        settings.setValue(QString("setup.thr[%1]").arg(i), setup.thr[i]);
        settings.setValue(QString("setup.trig[%1]").arg(i), bool(setup.trig[i]));
    }
    settings.setValue("mode", setup.mode);
    settings.setValue("readcellnumber", setup.readcellnumber);
    settings.setValue("writecellnumber", setup.writecellnumber);
    settings.setValue("trigdelay", setup.trigdelay);
    settings.endGroup();
}


void MainWindow::on_buttonBoxHostName_clicked(QAbstractButton* button)
{
    switch(ui->buttonBoxHostName->standardButton(button)) {
    case QDialogButtonBox::Apply :
        setBoardHostName();
        break;
    default:;
    }
}

void MainWindow::setBoardHostName()
{
    dominoDevice->setHostAddress(QHostAddress(ui->lineEditHostName->text()));
}

void MainWindow::on_pushButtonResetSettings_clicked()
{
    resetSettings();
    restoreFormValues();
}
