//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "mregdevice.h"

#include <QDebug>

#include "regio/crc8.h"

namespace mlink {

DeviceAttr::DeviceAttr()
    :
    device_id(0),
    id_chip_present(false),
    device_serial(0),
    temp(0),
    clock(0),
    fw_ver(0),
    fw_rev(0)
{
}

QString DeviceAttr::getFirmwareVersionStr() const
{
    int ver_major = (fw_ver >> 8) & 0xFF;
    int ver_minor = (fw_ver) & 0xFF;
    if ((fw_ver & 0xFFFF) == 0xFFFF) {
        ver_major = 0;
        ver_minor = 0;
    }
    return QString("%1.%2.%3").arg(ver_major).arg(ver_minor).arg(fw_rev);
}

MregDevice::MregDevice(RegIOBase &p, int address)
    : MlinkDevice(p, address)
{
}

MregDeviceBpm::MregDeviceBpm(RegIOBase &p, int address)
    : MregDevice(p, address)
{
}

MregDeviceAdcm::MregDeviceAdcm(RegIOBase &p, int address)
    : MregDevice(p, address)
{
}

quint64 MregDeviceBpm::readOneWireId()
{
    if (!hardwareAccessAllowed()) return 0;
    std::vector<quint16> v = regReadBlk(REG64_SERIAL_NUM, 0x4);
/*
    std::vector<quint16> v;
    v.push_back(regRead(REG_SERIAL_NUM));
    v.push_back(regRead(REG_SERIAL_NUM+1));
    v.push_back(regRead(REG_SERIAL_NUM+2));
    v.push_back(regRead(REG_SERIAL_NUM+3));
*/
    std::vector<quint16> c(8);
    bool all_zeroes = true;
    bool all_ones = true;
    for(int i=0; i<8; i++) {
        c[i] = (v[i/2] >> ((i%2)?8:0)) & 0xFF;
        if (c[i] != 0) all_zeroes = false;
        if (c[i] != 0xFF) all_ones = false;
    }
    if (all_zeroes || all_ones) {
        qWarning() << "No identification chip present";
        return 0;
    }

    //  for(size_t i=0; i<c.size(); i++)
    //	cout << hex << i << ": " << (unsigned int)c[i] << endl;

    const int testID = 0x28;
    unsigned char crc = 0;
    for(int i=0; i<7; i++)
        crc = crc8_dallas_maxim(c[i], crc);
    if (c[0] != testID) {
        qWarning() << QString("Unknown identification chip: read %1, expected %2").arg(c[0], 2, 16, QChar('0')).arg(testID, 2, 16, QChar('0'));
        return 0; // no-op
    }
    if (crc != c[7]) {
        qWarning() << QString("Identification chip CRC mismatch: read %1, calculated %2").arg(c[7], 2, 16, QChar('0')).arg((int)crc, 2, 16, QChar('0'));
        return 0;
    }
    quint64 id = quint64(c[1])
                 | (quint64(c[2]) << 8)
                 | (quint64(c[3]) << 16)
                 | (quint64(c[4]) << 24)
                 | (quint64(c[5]) << 32)
                 | (quint64(c[6]) << 40);
    attr.id_chip_present = true;
    attr.device_serial = id;
    return id;
}

quint64 MregDeviceAdcm::readOneWireId()
{
    if (!hardwareAccessAllowed()) return 0;
    quint16 id = regRead(ADCM16_SERIAL_ID_REG);
    attr.id_chip_present = true;
    attr.device_serial = id;
    return id;
}

bool MregDeviceBpm::readFirmwareId()
{
    if (!hardwareAccessAllowed()) return false;
    quint16 ver = regRead(REG_FW_VER);
    quint16 rev = regRead(REG_FW_REV);
    bool ver_present = (ver != 0 && ver != quint16(-1));
    bool rev_present = (rev != 0 && rev != quint16(-1));
    if (ver_present) attr.fw_ver = ver; else attr.fw_ver = 0;
    if (rev_present) attr.fw_rev = rev; else attr.fw_rev = 0;
    return ver_present && rev_present;
}

bool MregDeviceAdcm::readFirmwareId()
{
    if (!hardwareAccessAllowed()) return false;
    quint16 ver = regRead(ADCM16_FW_VER_REG);
    quint16 rev = regRead(ADCM16_FW_REV_REG);
    bool ver_present = (ver != 0 && ver != quint16(-1));
    bool rev_present = (rev != 0 && rev != quint16(-1));
    if (ver_present) attr.fw_ver = ver;
    if (rev_present) attr.fw_rev = rev;
    return ver_present && rev_present;
}

void MregDevice::print() const
{
  qDebug() << QString("Device ID %1, Firmware %2, Serial Number %3, Address %4, Temp %5")
          .arg(getDeviceId(), 4, 16, QChar('0'))
          .arg(attr.getFirmwareVersionStr())
          .arg(getDeviceSerial(), 8, 16, QChar('0'))
          .arg(getAddress(), 4, 16, QChar('0'))
          .arg(attr.temp, 0, 'f', 1);
}

double MregDeviceBpm::readTemp()
{
    if (!hardwareAccessAllowed()) return 0;
    int rawTemp = regRead(REG_TEMPERATURE) & 0xFFF;
    if (rawTemp & 0x800) rawTemp = -(rawTemp&0x7FF);
    attr.temp = 0.0625 * rawTemp;
    return attr.temp;
}

double MregDeviceAdcm::readTemp()
{
    if (!hardwareAccessAllowed()) return 0;
    int rawTemp = regRead(ADCM16_TEMPERATURE_REG) & 0xFFF;
    if (rawTemp & 0x800) rawTemp = -(rawTemp&0x7FF);
    attr.temp = 0.0625 * rawTemp;
    return attr.temp;
}

quint64 MregDeviceBpm::readTimestampCurrent()
{
    if (!hardwareAccessAllowed()) return 0;
    return regRead64(REG64_CURRENT_TIMESTAMP);
}

quint64 MregDeviceBpm::readTimestampRun()
{
    if (!hardwareAccessAllowed()) return 0;
    return regRead64(REG64_CNT_NCU);
}

quint64 MregDeviceBpm::readTimestampTrig()
{
    if (!hardwareAccessAllowed()) return 0;
    return regRead64(REG64_CNT_KCU);
}

quint64 MregDeviceAdcm::readTimestampCurrent()
{
    if (!hardwareAccessAllowed()) return 0;
    return regRead64(ADCM16_TS_READ_REG64);
}

quint64 MregDeviceAdcm::readTimestampRun()
{
    if (!hardwareAccessAllowed()) return 0;
    return regRead64(ADCM16_TS_READ_A);
}

quint64 MregDeviceAdcm::readTimestampTrig()
{
    if (!hardwareAccessAllowed()) return 0;
    return regRead64(ADCM16_TS_READ_B);
}


quint16 MregDevice::readDeviceId()
{
    quint16 id = regRead(REG_DEVICE_ID);
    qDebug() << QString("Read reg %1: %2").arg(REG_DEVICE_ID, 4, 16, QChar('0')).arg(id, 4, 16, QChar('0'));
    id &= 0xFF00;
    attr.device_id = id;
    return id;
}

//quint16 MregDeviceBpm::readDeviceId()
//{
//    quint16 id = regRead(REG_DEVICE_ID);
//    attr.device_id = id;
//    return id;
//}

//quint16 MregDeviceAdcm::readDeviceId()
//{
//    quint16 id = regRead(ADCM16_STATUS_REG) & 0xFF00;
//    attr.device_id = id;
//    return id;
//}

void MregDevice::detect()
{
    onlineState = false;
    if (!enableState) return;
//    qDebug() << QString("MregDevice: Trying address %1").arg(getAddress(), 4, 16, QChar('0'));
    try {
        quint16 id = readDeviceId();
        bool id_present = (id != 0) && ((id&0xFFFF) != 0xFFFF);
        if (!id_present) {
            qWarning() << QString("No device ID found");
            return;
        }
        onlineState = id_present;
        readOneWireId();
        readFirmwareId();
        readTemp();
    }

    catch (std::exception &e) {
        qWarning() << QString("MregDevice::detect failed: %1").arg(e.what());
    }

    if (isOnline())
        print();
}

} // namespace mlink
