//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "qtmregdevice.h"
#include "regio/regiomlink.h"
#include "regio/mregdevice.h"
#include "regio/mlinkdevice.h"

#include <QTimer>
#include <stdexcept>

const int onlineStatusPollTimeoutMs = 1000; // milliseconds
const int offlineStatusPollTimeoutMs = 5000; // milliseconds

QtMregDevice::QtMregDevice(QObject *parent) :
    QObject(parent),
    link(new mlink::RegIOMLink(this)),
    mregDevice(createMregDevice(*link, 0xFEFE, DEVICE_ID_AnyDevice)),
    timerStatusPoll(new QTimer(this)),
    currentTemp(0)
{
    connectedState = false;
    connect(timerStatusPoll, SIGNAL(timeout()), this, SLOT(statusPoll()));
    connect(this, SIGNAL(deviceConnected()), SLOT(on_deviceConnected()));
    connect(this, SIGNAL(deviceDisconnected()), SLOT(on_deviceDisconnected()));
    timerStatusPoll->setSingleShot(true);
//    timerStatusPoll->setInterval(statusPollTimeoutMs);
    timerStatusPoll->start(offlineStatusPollTimeoutMs);
}

QtMregDevice::~QtMregDevice()
{
    delete timerStatusPoll;
//    delete mregDevice;
    delete link;
}

mlink::MregDevice *QtMregDevice::createMregDevice(mlink::RegIOBase &link, int address, int id)
{
    switch (id) {
    case DEVICE_ID_TEST:      return new mlink::MregDeviceBpm(link, address);
    case DEVICE_ID_ADC64:     return new mlink::MregDeviceAdcm(link, address);
    case DEVICE_ID_ADC8E:     return new mlink::MregDeviceAdcm(link, address);
    case DEVICE_ID_8CADC:     return new mlink::MregDeviceAdcm(link, address);
    case DEVICE_ID_ADCM16:    return new mlink::MregDeviceAdcm(link, address);
    case DEVICE_ID_ADC16LTCM: return new mlink::MregDeviceAdcm(link, address);
    case DEVICE_ID_ADC16LTCS: return new mlink::MregDeviceAdcm(link, address);
    case DEVICE_ID_ADCM163:   return new mlink::MregDeviceAdcm(link, address);
    case DEVICE_ID_BPM4100:   return new mlink::MregDeviceBpm(link, address);
    case DEVICE_ID_EVADC:     return new mlink::MregDeviceBpm(link, address);
    case DEVICE_ID_NUCLORF:   return new mlink::MregDeviceBpm(link, address);
    case DEVICE_ID_BRIC:      return new mlink::MregDeviceAdcm(link, address);
    case DEVICE_ID_MSC8E:     return new mlink::MregDeviceBpm(link, address);
    case DEVICE_ID_ADCM164:   return new mlink::MregDeviceAdcm(link, address);
    case DEVICE_ID_ADC8BEDRS: return new mlink::MregDeviceBpm(link, address);
    default:
        return new mlink::MregDeviceDummy(link, address);
    }
}

QtMregDevice::DeviceId QtMregDevice::getDeviceId() const
{
    return static_cast<DeviceId>(mregDevice->getAttr().device_id);
}

quint64 QtMregDevice::getDeviceSerialNumber() const
{
    return static_cast<DeviceId>(mregDevice->getAttr().device_serial);
}

QString QtMregDevice::getFirmwareVersionStr() const
{
    return mregDevice->getAttr().getFirmwareVersionStr();
}

QString QtMregDevice::getDeviceIdStr() const
{
    DeviceId id = getDeviceId();
    QString str = QString("Unknown device %1").arg(id, 4, 16, QChar('0'));
    switch (id) {
    case QtMregDevice::DEVICE_ID_TEST: str = "TEST"; break;
    case QtMregDevice::DEVICE_ID_ADC64: str = "ADC64"; break;
    case QtMregDevice::DEVICE_ID_ADC8E: str = "ADC8E"; break;
    case QtMregDevice::DEVICE_ID_8CADC: str = "8CADC"; break;
    case QtMregDevice::DEVICE_ID_ADCM16: str = "ADCM16"; break;
    case QtMregDevice::DEVICE_ID_ADC16LTCM: str = "ADC16LTCM"; break;
    case QtMregDevice::DEVICE_ID_ADC16LTCS: str = "ADC16LTCS"; break;
    case QtMregDevice::DEVICE_ID_ADCM163: str = "ADCM163"; break;
    case QtMregDevice::DEVICE_ID_BPM4100: str = "BPM4100"; break;
    case QtMregDevice::DEVICE_ID_EVADC: str = "EVADC"; break;
    case QtMregDevice::DEVICE_ID_NUCLORF: str = "NUCLORF"; break;
    case QtMregDevice::DEVICE_ID_BRIC: str = "BRIC"; break;
    case QtMregDevice::DEVICE_ID_MSC8E: str = "MSC8E"; break;
    case QtMregDevice::DEVICE_ID_ADCM164: str = "ADCM164"; break;
    case QtMregDevice::DEVICE_ID_ADC8BEDRS: str = "ADC8BEDRS"; break;
    case QtMregDevice::DEVICE_ID_AnyDevice: break;
    }
    QString fw_id = mregDevice->getAttr().getFirmwareVersionStr();
    if (!fw_id.isEmpty())
        str += " "+fw_id;
    return str;
}

double QtMregDevice::getTemperature() const
{
    return mregDevice->getAttr().temp;
}

void QtMregDevice::setHostAddress(const QHostAddress &address)
{
    if (connectedState && (address == link->getHostAddress()))
        return;
    if (connectedState) {
        mregDevice->disable();
        connectedState = false;
        Q_EMIT deviceDisconnected();
        Q_EMIT setStatus(QString("Disconnected"));
    }
    link->setHostAddress(address);
    detect();
}

void QtMregDevice::detect()
{
    QString offlineReason = "Disconnected";
    if (connectedState) {
        Q_EMIT setStatus(QString("Reconnecting to %1...").arg(link->getHostAddress().toString()));
    } else {
        Q_EMIT setStatus(QString("Connecting to %1...").arg(link->getHostAddress().toString()));
    }
    mregDevice->enable();
    quint16 old_id = mregDevice->getAttr().device_id;
    quint16 id = 0;
    try {
        id = mregDevice->readDeviceId();
    }
    catch (...) {}
//    qDebug() << QString("Device ID:%1").arg(id, 4, 16, QChar('0'));
    if (id == 0 || id == 0xFFFF) {
        mregDevice->disable();
        QString msg = QString("No device ID found");
        qCritical() << msg;
        offlineReason = msg;
        Q_EMIT setStatus(msg);
    } else {
        if (old_id != id) {
//            qDebug() << QString("Device ID changed: %1 to %2, recreating device").arg(old_id, 4, 16, QChar('0')).arg(id, 4, 16, QChar('0'));
            mregDevice = QSharedPointer<mlink::MregDevice>(createMregDevice(*link, 0xFEFE, id));
            mregDevice->enable();
        }
        mregDevice->detect();
        if (mregDevice->isOnline()) {
            quint16 id = mregDevice->getAttr().device_id;
            bool supportedDevice = (supportedDeviceId.contains(DEVICE_ID_AnyDevice))
                    || supportedDeviceId.contains(id);
            if (!supportedDevice) {
                mregDevice->disable();
                QString msg = QString("Unknown device %1").arg(id, 4, 16, QChar('0'));
                qWarning() << msg;
                offlineReason = msg;
                Q_EMIT setStatus(msg);
            } else {
                mregDevice->enable();
            }
        }
    }
    bool newState = isConnected();
    if (newState && !connectedState) {
        Q_EMIT deviceConnected();
        Q_EMIT setStatus(QString("Online"));
    }
    if (!newState && connectedState) {
        Q_EMIT deviceDisconnected();
        Q_EMIT setStatus(QString(offlineReason));
    }
    if (newState && connectedState) {
        Q_EMIT setStatus(QString("Online"));
    }
    if (!newState && !connectedState) {
        Q_EMIT setStatus(QString(offlineReason));
    }
    connectedState = newState;
}

bool QtMregDevice::isConnected() const
{
    return mregDevice->isOnline() && mregDevice->isEnabled();
}

double QtMregDevice::readTemp()
{
    double temp = 0;
    try {
        temp = mregDevice->readTemp();
    }
    catch (std::runtime_error &e) {
        Q_EMIT setStatus(QString("Communication error: %1").arg(e.what()));
    }

    if (temp != currentTemp) {
        Q_EMIT TempChanged(temp);
        currentTemp = temp;
    }
    return temp;
}

quint64 QtMregDevice::readTimestampCurrent()
{
    return mregDevice->readTimestampCurrent();
}

quint64 QtMregDevice::readTimestampRun()
{
    return mregDevice->readTimestampRun();
}

quint64 QtMregDevice::readTimestampTrig()
{
    return mregDevice->readTimestampTrig();
}

void QtMregDevice::enable()
{
    mregDevice->enable();
}

void QtMregDevice::disable()
{
    mregDevice->disable();
}

void QtMregDevice::statusPoll()
{
//    if (!mregDevice->isEnabled())
//        goto finish;
    if (!mregDevice->isOnline()) {
        //        Q_EMIT setStatus(QString("Offline"));
        //qDebug() << "re-detect";
        detect(); // re-detect
        //        goto finish;
    }
    if (mregDevice->isOnline()) {
        readTemp();
    }

//finish:
    // restart timer
    if (mregDevice->isOnline())
        timerStatusPoll->start(onlineStatusPollTimeoutMs);
    else
        timerStatusPoll->start(offlineStatusPollTimeoutMs);
}

void QtMregDevice::on_deviceConnected()
{
    timerStatusPoll->start(0);
}

void QtMregDevice::on_deviceDisconnected()
{
    timerStatusPoll->start(offlineStatusPollTimeoutMs);
}

/// read 16-bit register
quint16 QtMregDevice::regRead(int regNumber)
{
//    Q_ASSERT(isConnected());
    try {
//        Q_EMIT setStatus(QString("regRead %1").arg(regNumber));
        return mregDevice->regRead(regNumber);
//        Q_EMIT setStatus(QString("Ok"));
    }
    catch (std::runtime_error &e) {
        Q_EMIT setStatus(QString("Communication error: %1").arg(e.what()));
        return 0;
    }
}

/// write 16-bit register
void QtMregDevice::regWrite(int regNumber, quint16 data)
{
//    Q_ASSERT(isConnected());
    try {
//        Q_EMIT setStatus(QString("regWrite %1 %2").arg(regNumber).arg(data));
        mregDevice->regWrite(regNumber, data);
//        Q_EMIT setStatus(QString("Ok"));
    }
    catch (std::runtime_error &e) {
        Q_EMIT setStatus(QString("Communication error: %1").arg(e.what()));
    }
}

//void QtMregDevice::regWrite(const std::map<quint16, quint16> &data)
//{
//    Q_ASSERT(isConnected());
//    try {
//        Q_EMIT setStatus(QString("regWriteMap"));
//        mregDevice->regWriteBlk(data);
//        Q_EMIT setStatus(QString("Ok"));
//    }
//    catch (std::runtime_error &e) {
//        Q_EMIT setStatus(QString("Communication error: %1").arg(e.what()));
//    }
//}

/// read 32-bit register
quint32 QtMregDevice::regRead32(int regNumber)
{
//    Q_ASSERT(isConnected());
    try {
//        Q_EMIT setStatus(QString("regRead32 %1").arg(regNumber));
        return mregDevice->regRead32(regNumber);
//        Q_EMIT setStatus(QString("Ok"));
    }
    catch (std::runtime_error &e) {
        Q_EMIT setStatus(QString("Communication error: %1").arg(e.what()));
        return 0;
    }
}

/// write 32-bit register
void QtMregDevice::regWrite32(int regNumber, quint32 data)
{
//    Q_ASSERT(isConnected());
    try {
//        Q_EMIT setStatus(QString("regWrite32 %1 %2").arg(regNumber).arg(data));
        mregDevice->regWrite32(regNumber, data);
//        Q_EMIT setStatus(QString("Ok"));
    }
    catch (std::runtime_error &e) {
        Q_EMIT setStatus(QString("Communication error: %1").arg(e.what()));
    }
}

/// read 64-bit register
quint64 QtMregDevice::regRead64(int regNumber)
{
//    Q_ASSERT(isConnected());
    try {
//        Q_EMIT setStatus(QString("regRead64 %1").arg(regNumber));
        return mregDevice->regRead64(regNumber);
//        Q_EMIT setStatus(QString("Ok"));
    }
    catch (std::runtime_error &e) {
        Q_EMIT setStatus(QString("Communication error: %1").arg(e.what()));
        return 0;
    }
}

/// write 64-bit register
void QtMregDevice::regWrite64(int regNumber, quint64 data)
{
//    Q_ASSERT(isConnected());
    try {
//        Q_EMIT setStatus(QString("regWrite64 %1 %2").arg(regNumber).arg(data));
        mregDevice->regWrite64(regNumber, data);
//        Q_EMIT setStatus(QString("Ok"));
    }
    catch (std::runtime_error &e) {
        Q_EMIT setStatus(QString("Communication error: %1").arg(e.what()));
    }
}

/// evaluate register operations
mlink::RegOpVector QtMregDevice::regOpExec(const mlink::RegOpVector &data)
{
    mlink::RegOpVector result;
//    Q_ASSERT(isConnected());
    try {
//        Q_EMIT setStatus(QString("regOpExec %1").arg(data.size()));
        result = mregDevice->regOpExec(data);
//        Q_EMIT setStatus(QString("Ok"));
    }
    catch (std::runtime_error &e) {
        Q_EMIT setStatus(QString("Communication error: %1").arg(e.what()));
    }
    return result;
}

std::vector<quint32> QtMregDevice::memReadBlk(int startWord, size_t wordCount)
{
//    Q_ASSERT(isConnected());
    try {
//        Q_EMIT setStatus(QString("memReadBlk %1 %5").arg(startWord).arg(wordCount));
        return mregDevice->memReadBlk(startWord, wordCount);
//        Q_EMIT setStatus(QString("Ok"));
    }
    catch (std::runtime_error &e) {
        Q_EMIT setStatus(QString("Communication error: %1").arg(e.what()));
        return std::vector<quint32>();
    }
}

quint32 QtMregDevice::memRead(int wordNumber)
{
//    Q_ASSERT(isConnected());
    try {
//        Q_EMIT setStatus(QString("memWrite %1 %5").arg(wordNumber).arg(data));
        return mregDevice->memRead(wordNumber);
//        Q_EMIT setStatus(QString("Ok"));
    }
    catch (std::runtime_error &e) {
        Q_EMIT setStatus(QString("Communication error: %1").arg(e.what()));
        return 0;
    }
}

void QtMregDevice::memWrite(int wordNumber, quint32 data)
{
//    Q_ASSERT(isConnected());
    try {
//        Q_EMIT setStatus(QString("memWrite %1 %5").arg(wordNumber).arg(data));
        mregDevice->memWrite(wordNumber, data);
//        Q_EMIT setStatus(QString("Ok"));
    }
    catch (std::runtime_error &e) {
        Q_EMIT setStatus(QString("Communication error: %1").arg(e.what()));
    }
}

void QtMregDevice::memWriteBlk(int wordNumber, const std::vector<quint32> &data)
{
//    Q_ASSERT(isConnected());
    try {
//        Q_EMIT setStatus(QString("memWriteBlk %1 %5").arg(wordNumber).arg(data.size()));
        mregDevice->memWriteBlk(wordNumber, data);
//        Q_EMIT setStatus(QString("Ok"));
    }
    catch (std::runtime_error &e) {
        Q_EMIT setStatus(QString("Communication error: %1").arg(e.what()));
    }
}
