//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#ifndef QTMREGDEVICE_H
#define QTMREGDEVICE_H

#include <QObject>
#include <QMap>
#include <QSharedPointer>
#include <vector>

namespace mlink {
class MregDevice;
class RegIOBase;
class RegOpVector;
}

class QHostAddress;
class QTimer;

class QtMregDevice : public QObject
{
    Q_OBJECT
public:
    typedef enum {
        DEVICE_ID_TEST      = 0x5500,
        DEVICE_ID_8CADC     = 0xB300,
        DEVICE_ID_ADCM16    = 0xB400,
        DEVICE_ID_ADC16LTCM = 0xB500,
        DEVICE_ID_ADC16LTCS = 0xB501,
        DEVICE_ID_ADCM163   = 0xB600,
        DEVICE_ID_BPM4100   = 0xB700,
        DEVICE_ID_ADC8E     = 0xB800,
        DEVICE_ID_EVADC     = 0xB900,
        DEVICE_ID_ADC64     = 0xBA00,
        DEVICE_ID_NUCLORF   = 0xBB00,
        DEVICE_ID_BRIC      = 0xBC00,
        DEVICE_ID_MSC8E     = 0xBD00,
        DEVICE_ID_ADCM164   = 0xBE00,
        DEVICE_ID_ADC8BEDRS = 0xBF00,
        DEVICE_ID_AnyDevice = 0xFFFF
    } DeviceId;

    explicit QtMregDevice(QObject *parent = 0);
    virtual ~QtMregDevice();
    void addSupportedDeviceId(int id) {supportedDeviceId[id] = true;}
    void detect();
    bool isConnected() const;
    DeviceId getDeviceId() const;
    quint64 getDeviceSerialNumber() const;
    QString getDeviceIdStr() const;
    QString getFirmwareVersionStr() const;
    double getTemperature() const;

    /// read 16-bit register
    quint16 regRead(int regNumber);

    /// write 16-bit register
    void regWrite(int regNumber, quint16 data);

//    void regWrite(const QMap<quint16, quint16> &data);

    /// read 32-bit register
    quint32 regRead32(int regNumber);

    /// write 32-bit register
    void regWrite32(int regNumber, quint32 data);

    /// read 64-bit register
    quint64 regRead64(int regNumber);

    /// write 64-bit register
    void regWrite64(int regNumber, quint64 data);

    /// evaluate register operations
    mlink::RegOpVector regOpExec(const mlink::RegOpVector &data);

    /// read 32-bit word from RAM at word offset wordNumber
    quint32 memRead(int wordNumber);

    /// write 32-bit word to RAM at word offset wordNumber
    void memWrite(int wordNumber, quint32 data);

    /// read wordCount 32-bit words from RAM starting from startWord offset
    std::vector<quint32> memReadBlk(int startWord, size_t wordCount);

    /// write 32-bit words to RAM starting from startWord offset
    void memWriteBlk(int wordNumber, const std::vector<quint32> &data);

Q_SIGNALS:
    void deviceConnected();
    void deviceDisconnected();
    void setStatus(const QString &);
    void TempChanged(double);

private Q_SLOTS:
    void statusPoll();
    void on_deviceConnected();
    void on_deviceDisconnected();

public Q_SLOTS:
    void setHostAddress(const QHostAddress &address);

protected:
    mlink::MregDevice *createMregDevice(mlink::RegIOBase &p, int address, int id);
    mlink::RegIOBase * const link;
    double readTemp();
    quint64 readTimestampCurrent();
    quint64 readTimestampRun();
    quint64 readTimestampTrig();
    void enable();
    void disable();

protected:
private:
    QSharedPointer<mlink::MregDevice> mregDevice; // dev must be after link
    QTimer * const timerStatusPoll;
    bool connectedState;
    QMap<int, bool> supportedDeviceId;
    double currentTemp;
};

#endif // QTMREGDEVICE_H
