
#include "regiomlink.h"

#include <QDebug>
#include <string>
#include <sstream>
#include <iomanip>
#include <cstring>
#include <stdexcept>

#include <QByteArray>
#include <QBuffer>
#include <QUdpSocket>

#include "regio/regiopacket.h"
#include "mlink/mlink.h"
#include "mlink/frame_types.h"
#include "mlink/mlink_frame_header.h"
#include "util/gettimeofday.h"
#include "regio/TextFormatter.h"

#include <zlib.h>

namespace mlink {

const int send_timeout = 1000; // use -1 to disable wait, return immediately
const int receive_timeout = 1000; // ms

RegIOMLink::RegIOMLink(QObject *parent, const QHostAddress &address)
    : RegIOBase(parent, address), socket(new QUdpSocket(parent))
{
    bufSizeLimit = 10000; // maximum number of packets
//    socket->connectToHost(address, ML_UDP_PORT);
//    if (socket->waitForConnected(1000))
//        qDebug() << "Connected to " << socket->peerAddress();
//    else {
//        qWarning() << "Connect failed: " << socket->errorString();
//        throw std::runtime_error("connect failed");
//    }
    if (!socket->bind(0, QUdpSocket::DontShareAddress/*, QUdpSocket::ReuseAddressHint*/)) {
        QString str = QString("Can't bind to UDP port %1: %2")
                      .arg(ML_UDP_PORT)
                      .arg(socket->errorString());
        qWarning() << str;
        throw std::runtime_error(str.toStdString().c_str());
    }
    connect(socket, SIGNAL(readyRead()), this, SLOT(readSocketData()));
}

RegIOMLink::~RegIOMLink()
{
    delete socket;
}

//std::string RegIOMLink::defaultDevName()
//{
//    return std::string(ML_DEVICE_NAME)+std::string("0");
//}

void RegIOMLink::reset()
{
    RegIOBase::reset();
}


double RegIOMLink::getDeviceThroughput() const
{
    return 160e6;
}

unsigned long RegIOMLink::getMaxDataSize() const
{
    return ML_MAX_FRAME_DATA_WORDS;
}

/// Serialize packet
void RegIOMLink::rawSend(const RegIoPacket &pkt)
{
//    int packetSize = 4 * (ML_FRAME_HEADER_WORDS + ML_FRAME_TRAILER_WORDS);
    std::vector<quint16> v;
    v.push_back(pkt.type);
    v.push_back(ML_FRAME_SYNC);
    v.push_back(pkt.seq);
    v.push_back((ML_FRAME_HEADER_WORDS + ML_FRAME_TRAILER_WORDS) + pkt.data.size());
    v.push_back(pkt.src);
    v.push_back(pkt.dst);
    QByteArray buf(reinterpret_cast<const char *>(&v[0]), 2 * v.size());
    const int dataSize = pkt.data.size();
    QByteArray bufData(reinterpret_cast<const char *>(&pkt.data[0]), sizeof(PacketRawDataType) * dataSize);
    buf.append(bufData);
    quint32 checksum = crc32(0, reinterpret_cast<const Bytef *>(buf.data()), buf.size());
    buf.append(reinterpret_cast<const char *>(&checksum), 4);
    if (deviceAddress.isNull()) {
        qWarning() << "device address not set";
        return;
    }
    qint64 rc = socket->writeDatagram(buf, deviceAddress, ML_UDP_PORT);
//    qint64 rc = socket->write(buf);
//    {
//        //  printf("tx: size=%d\n", data.data.size());
//        std::ostringstream ost;
//        ost << "RegIOMLink::rawSend() " << data;
//        log_debug(0, ost.str());
//    }
    if (rc != buf.size()) {
        throw std::runtime_error(QString("Frame send failed: %1").arg(socket->errorString()).toStdString());
    }
//    socket->waitForBytesWritten(1000);
//    socket->waitForReadyRead(1000);
    socket->flush();
}

/// Deserialize packet
void RegIOMLink::processPacket(const QByteArray &buf)
{
    if ((buf.size() < ML_MIN_RAW_FRAME_OCTETS) || (buf.size() < ML_MIN_RAW_FRAME_OCTETS)) {
        qWarning() << "Bad frame dropped: size " << buf.size();
        for(int i=0; i<buf.size(); i++)
            qDebug() << QString("%1").arg((unsigned char)(buf[i]), 2, 16, QChar('0'));
        return;
    }
    const quint16 *type = reinterpret_cast<const quint16 *>(&buf.constData()[0]);
    const quint16 *sync = reinterpret_cast<const quint16 *>(&buf.constData()[2]);
    const quint16 *seq = reinterpret_cast<const quint16 *>(&buf.constData()[4]);
    const quint16 *len = reinterpret_cast<const quint16 *>(&buf.constData()[6]);
    const quint16 *src = reinterpret_cast<const quint16 *>(&buf.constData()[8]);
    const quint16 *dst = reinterpret_cast<const quint16 *>(&buf.constData()[10]);
    const quint16 *data = reinterpret_cast<const quint16 *>(&buf.constData()[12]);
    if (*sync != ML_FRAME_SYNC) {
        qWarning() << "Bad frame dropped: sync " << *sync;
        return;
    }
    if (*len < ML_MIN_RAW_FRAME_WORDS || *len > ML_MAX_RAW_FRAME_WORDS || *len*4 != buf.size()) {
        qWarning() << "Bad frame dropped: len " << *len << ", buf.size " << buf.size();
        return;
    }
    if (packetTypeIsData(*type)) {
//        qDebug() << "Received data packet " << buf.size() << " bytes";
        Q_EMIT rawDataReceived(QByteArray(&buf.constData()[12], buf.size() - ML_MIN_RAW_FRAME_OCTETS));
    } else {
//        qDebug() << "Received ctrl packet seq" << *seq;
        int bufSize = buf.size();
        int headerSize = 4 * (ML_FRAME_HEADER_WORDS + ML_FRAME_TRAILER_WORDS);
        int dataSize = bufSize - headerSize;
        PacketData pd;
        pd.resize(dataSize / sizeof(PacketRawDataType));
        memcpy(&pd[0], data, dataSize);
        RegIoPacket pkt(*type, *seq, *src, *dst, pd);
        rxBufCtrl.push_back(pkt);
//        qDebug() << "processPacket: rxBufCtrl=" << rxBufCtrl;
    }
}

void RegIOMLink::readSocketData()
{
//    qWarning() << "RegIOMLink::readSocketData()";
    while (socket->hasPendingDatagrams()) {
        QByteArray datagram;
        datagram.resize(socket->pendingDatagramSize());
        QHostAddress sender;
        quint16 senderPort;

        socket->readDatagram(datagram.data(), datagram.size(),
                                &sender, &senderPort);
        processPacket(datagram);
    }
}

} // namespace mlink
