
#include <QApplication>
#include <QEvent>
#include <QMouseEvent>
#include <QWhatsThis>
#include <QPainter>
#include <QDebug>
#include <qwt/qwt_plot.h>
#include <qwt/qwt_symbol.h>
#include <qwt/qwt_scale_map.h>
#include <qwt/qwt_plot_canvas.h>
#include <qwt/qwt_plot_curve.h>

#include "qxw/QwtPlotEdit.h"


QwtArray<double> insertAt(const QwtArray<double> &data, int index, double value)
{
    const int dataSize = data.size();
    QwtArray<double> dataNew(dataSize+1);
    if (index < 0 || index > dataSize) index = dataSize; // add point to end() if position not specified (-1)
    for(int i=0; i<index; i++)
    {
        dataNew[i] = data[i];
    }
    dataNew[index] = value;
    for(int i=index; i<dataSize; i++)
    {
        dataNew[i+1] = data[i];
    }
    return dataNew;
}

QwtArray<double> removeAt(const QwtArray<double> &data, int index)
{
    const int dataSize = data.size();
    QwtArray<double> dataNew(dataSize-1);
    for(int i=0; i<index; i++)
    {
        dataNew[i] = data[i];
    }
    for(int i=index; i<dataSize-1; i++)
    {
        dataNew[i] = data[i+1];
    }
    return dataNew;
}

MoveCommand::MoveCommand(QwtPlotEdit *picker, QwtPlotCurve *curve, int index, double dx, double dy)
    : myPicker(picker), myCurve(curve), myIndex(index), myDx(dx), myDy(dy)
{
}

QString MoveCommand::myText() const
{
    return QObject::tr("move point %1 by %2:%3").arg(myIndex).arg(myDx).arg(myDy);
}

bool MoveCommand::mergeWith(const QUndoCommand *command)
{
    const MoveCommand *moveCommand = static_cast<const MoveCommand *>(command);
    const QwtPlotCurve *c = moveCommand->myCurve;
    const int index = moveCommand->myIndex;
    if (c!=myCurve || index!=myIndex) {
        qDebug() << "no merge";
        return false;
    }
    myDx += moveCommand->myDx;
    myDy += moveCommand->myDy;
    qDebug() << "merge";
    setText(myText());
    return true;
}

void MoveCommand::redo()
{
    doMove(true);
}

void MoveCommand::undo()
{
    doMove(false);
}

void MoveCommand::doMove(bool forward)
{
    if (!myCurve) return; // no curve
    const int dataSize = myCurve->dataSize();
    if (myIndex < 0 || myIndex >= dataSize) return; // out of range
    const QwtArray<double> xDataOrig = reinterpret_cast<const QwtArrayData &>(myCurve->data()).xData();
    const QwtArray<double> yDataOrig = reinterpret_cast<const QwtArrayData &>(myCurve->data()).yData();
    QwtArray<double> xData(xDataOrig);
    QwtArray<double> yData(yDataOrig);
    if (forward) {
        qDebug() << "doMove(forward) " << myDx << ":" << myDy;
        xData[myIndex] += myDx;
        yData[myIndex] += myDy;
    } else {
        qDebug() << "doMove(reverse) " << -myDx << ":" << -myDy;
        xData[myIndex] -= myDx;
        yData[myIndex] -= myDy;
    }


    myCurve->setData(xData, yData);

    QwtPlot *plot = myCurve->plot();
    if (plot) plot->replot();
    //    myPicker->updateSelection();
}

AddCommand::AddCommand(QwtPlotEdit *picker, QwtPlotCurve *curve, int index, double px, double py)
    : myPicker(picker), myCurve(curve), myIndex(index), myPx(px), myPy(py)
{
}

QString AddCommand::myText() const
{
    return QObject::tr("add point %1 at %2:%3").arg(myIndex).arg(myPx).arg(myPy);
}

void AddCommand::redo()
{
    doAdd();
}

void AddCommand::undo()
{
    doRemove();
}

void AddCommand::doAddRemove(int action)
{
    if (!myCurve) return; // no curve
    const int dataSize = myCurve->dataSize();
    if (myIndex < 0 || myIndex > dataSize) return; // out of range
    const QwtArray<double> xDataOrig = reinterpret_cast<const QwtArrayData &>(myCurve->data()).xData();
    const QwtArray<double> yDataOrig = reinterpret_cast<const QwtArrayData &>(myCurve->data()).yData();
    QwtArray<double> xData;
    QwtArray<double> yData;
    if (action == ActionAdd) {
        qDebug() << "add " << myIndex << " at " << myPx << ":" << myPy;
        xData = insertAt(xDataOrig, myIndex, myPx);
        yData = insertAt(yDataOrig, myIndex, myPy);
    } else {
        qDebug() << "remove " << myIndex;
        xData = removeAt(xDataOrig, myIndex);
        yData = removeAt(yDataOrig, myIndex);
    }
    myCurve->setData(xData, yData);

    QwtPlot *plot = myCurve->plot();
    if (plot) plot->replot();
    //    myPicker->updateSelection();
}

RemoveCommand::RemoveCommand(QwtPlotEdit *picker, QwtPlotCurve *curve, int index)
    : AddCommand(picker, curve, index, 0, 0)
{
}

QString RemoveCommand::myText() const
{
    return QObject::tr("remove point %1").arg(myIndex);
}

void RemoveCommand::redo()
{
    doRemove();
}

void RemoveCommand::undo()
{
    doAdd();
}

void QwtPlotEdit::redo()
{
    undoStack->redo();
}

void QwtPlotEdit::undo()
{
    undoStack->undo();
}

QwtPlotEdit::QwtPlotEdit(QwtPlot *plot, QUndoStack *stack):
        QObject(plot),
        pickerButton(Qt::LeftButton),
        undoStack(stack)
{
    QwtPlotCanvas *canvas = plot->canvas();

    canvas->installEventFilter(this);

    // We want the focus, but no focus rect. The
    // selected point will be highlighted instead.

#if QT_VERSION >= 0x040000
    canvas->setFocusPolicy(Qt::StrongFocus);
#ifndef QT_NO_CURSOR
    canvas->setCursor(Qt::PointingHandCursor);
#endif
#else
    canvas->setFocusPolicy(QWidget::StrongFocus);
#ifndef QT_NO_CURSOR
    canvas->setCursor(Qt::pointingHandCursor);
#endif
#endif
    canvas->setFocusIndicator(QwtPlotCanvas::ItemFocusIndicator);
    canvas->setFocus();

    QString text =
            tr("Pick and move points with left mouse button or with these keys")+":\n\n"+
            "- "+tr("Up")+":\t\t"+tr("Select next curve")+"\n"+
            "- "+tr("Down")+":\t\t"+tr("Select previous curve")+"\n"+
            "- "+tr("Left")+", ´-´:\t"+tr("Select next point")+"\n"+
            "- "+tr("Right")+", ´+´:\t"+tr("Select previous point")+"\n"+
            "- 7, 8, 9, 4, 6, 1, 2, 3:\t"+tr("Move selected point")+"\n"+
            "- "+tr("Insert")+":\t\t"+tr("Add new point")+"\n"+
            "- "+tr("Delete")+":\t\t"+tr("Remove selected point")+"\n";
#if QT_VERSION >= 0x040000
    canvas->setWhatsThis(text);
#else
    QWhatsThis::add(canvas, text);
#endif

    shiftCurveCursor(true);
}

void QwtPlotEdit::setMouseButton(int button, int buttonState)
{
    Q_UNUSED(buttonState);
    pickerButton = button;
}

bool QwtPlotEdit::event(QEvent *e)
{
    if ( e->type() == QEvent::User )
    {
        showCursor(true);
        return true;
    }
    return QObject::event(e);
}

bool QwtPlotEdit::eventFilter(QObject *object, QEvent *e)
{
    if ( object != (QObject *)plot()->canvas() )
        return false;

    switch(e->type())
    {
    case QEvent::FocusIn:
        {
            showCursor(true);
            //            undoStack->setActive(true);
            break;
        }
    case QEvent::FocusOut:
        {
            showCursor(false);
            //            undoStack->setActive(false);
            break;
        }
    case QEvent::Paint:
        {
            QApplication::postEvent(this, new QEvent(QEvent::User)); //MemLeak?
            break;
        }
    case QEvent::MouseButtonPress:
        {
            const QMouseEvent *me = (const QMouseEvent *)e;
            //            qDebug() << "press with buttons " << me->buttons();
            if (me->buttons() & pickerButton) {
                select(me->pos());
                return true;
            }
            break;
        }
    case QEvent::MouseMove:
        {
            const QMouseEvent *me = (const QMouseEvent *)e;
            //            qDebug() << "move with buttons " << me->buttons();
            if (me->buttons() & pickerButton) {
                move(me->pos());
                return true;
            }
            break;
        }
    case QEvent::KeyPress:
        {
            const int delta = 5;
            switch(((const QKeyEvent *)e)->key())
            {
            case Qt::Key_Up:
                shiftCurveCursor(true);
                return true;

            case Qt::Key_Down:
                shiftCurveCursor(false);
                return true;

            case Qt::Key_Right:
            case Qt::Key_Plus:
                if ( ! d_select.empty() )
                    shiftPointCursor(true);
                else
                    shiftCurveCursor(true);
                return true;

            case Qt::Key_Left:
            case Qt::Key_Minus:
                if ( ! d_select.empty() )
                    shiftPointCursor(false);
                else
                    shiftCurveCursor(true);
                return true;

                // The following keys represent a direction, they are
                // organized on the keyboard.

            case Qt::Key_1:
                moveBy(-delta, delta);
                break;
            case Qt::Key_2:
                moveBy(0, delta);
                break;
            case Qt::Key_3:
                moveBy(delta, delta);
                break;
            case Qt::Key_4:
                moveBy(-delta, 0);
                break;
            case Qt::Key_6:
                moveBy(delta, 0);
                break;
            case Qt::Key_7:
                moveBy(-delta, -delta);
                break;
            case Qt::Key_8:
                moveBy(0, -delta);
                break;
            case Qt::Key_9:
                moveBy(delta, -delta);
                break;
            case Qt::Key_Delete:
                deletePoint();
                break;
            default:
                break;
            }
        }
    default:
        break;
    }
    return QObject::eventFilter(object, e);
}

/*!
  Find the closest curve line segment for a specific position

  \param pos Position, where to look for the closest curve point
  \param dist If dist != NULL, closestSegment() returns the distance between
              the position and the clostest curve segment
  \return Index of the first point of closest segment, or -1 if none can be found
          ( f.e when the curve has no points )
  \note closestSegment() implements a dumb algorithm, that iterates
        over all points
  \note Linear interpolated curves only. No spline support.
*/

double distance2piece(double qx, double qy, double px1, double py1, double px2, double py2)
{
    double t;
    const double epsilon = 1e-14; // enough for display calculations
    if (((qx-px1)*(px2-px1)+(qy-py1)*(py2-py1))*
        ((qx-px2)*(px2-px1)+(qy-py2)*(py2-py1))>-epsilon)
    {
        t = qwtSqr(qx-px1)+qwtSqr(qy-py1);
        double w = qwtSqr(qx-px2)+qwtSqr(qy-py2);
        if (w<t) {t= w;}
    } else
        t= qwtSqr((qx-px1)*(py2-py1)-(qy-py1)*(px2-px1))/
           (qwtSqr(px2-px1)+qwtSqr(py2-py1));
    return sqrt(t);
}

int findClosestSegment(const QwtPlotCurve *c, const QPoint &pos, double *dist)
{
    if ( c->plot() == NULL || c->dataSize() <= 1 )
        return -1;

    const QwtScaleMap xMap = c->plot()->canvasMap(c->xAxis());
    const QwtScaleMap yMap = c->plot()->canvasMap(c->yAxis());

    int index = -1;
    double dmin = 1.0e10;

    // iterate over line segments
    for (int i=0; i < c->dataSize()-1; i++)
    {
        const double qx = pos.x();
        const double qy = pos.y();
        const double px1 = xMap.xTransform(c->x(i));
        const double py1 = yMap.xTransform(c->y(i));
        const double px2 = xMap.xTransform(c->x(i+1));
        const double py2 = yMap.xTransform(c->y(i+1));
        double f = distance2piece(qx, qy, px1, py1, px2, py2);
        if (f < dmin)
        {
            index = i;
            dmin = f;
        }
    }
    if ( dist )
        *dist = dmin;

    return index;
}

const double maxdist = 1e38;
CurveSelection::CurveSelection()
{
    clear();
}

void CurveSelection::clear()
{
    c = 0;
    dist = maxdist;
    index = -1;
    type = -1;
}

CurveSelection::CurveSelection(QwtPlotCurve *_c, double _dist, double _index, int _type)
    : c(_c), dist(_dist), index(_index), type(_type)
{
}

// Select the point at a position. If there is no point
// deselect the selected point

void QwtPlotEdit::select(const QPoint &pos)
{
    CurveSelection pointSelection, segmentSelection;
    double distPoint = maxdist;
    double distSeg = maxdist;

    const QwtPlotItemList& itmList = plot()->itemList();
    for ( QwtPlotItemIterator it = itmList.begin();
    it != itmList.end(); ++it )
    {
        if ( (*it)->rtti() == QwtPlotItem::Rtti_PlotCurve )
        {
            QwtPlotCurve *c = (QwtPlotCurve*)(*it);
            {

                double d;
                int idx = c->closestPoint(pos, &d);
                if ( d < distPoint )
                {
                    pointSelection = CurveSelection(c, d, idx, CurveSelection::Point);
                    distPoint = d;
                }
            }
            {
                double d;
                int idx = findClosestSegment(c, pos, &d);
                if ( d < distSeg )
                {
                    segmentSelection = CurveSelection(c, d, idx, CurveSelection::Segment);
                    distSeg = d;
                }
            }
        }
    }

    CurveSelection cur_select = CurveSelection();

    if ( !pointSelection.empty() && distPoint < 10 ) // 10 pixels tolerance
    {
        cur_select = pointSelection;
        qDebug() << "point";
    } else
        if ( !segmentSelection.empty() && distSeg < 10 ) // 10 pixels tolerance
        {
        qDebug() << "segment";
        cur_select = segmentSelection;
    }
    //    if (d_select.empty() && cur_select.empty()) return; // nothing to do
    if (cur_select.isSegment()) {
        cur_select.type = CurveSelection::Point;
        cur_select.index++;
        addPoint(cur_select.c, pos, cur_select.index);
        newSelection(cur_select);
    } else {
        newSelection(cur_select);
    }
}

// Add point with coordinates p to curve c at index
void QwtPlotEdit::addPoint(QwtPlotCurve *c, const QPoint &p, int index)
{
    //    if ( d_select.empty() )
    //        return;
    if (!c || !c->plot()) return;
    const QwtScaleMap xMap = c->plot()->canvasMap(c->xAxis());
    const QwtScaleMap yMap = c->plot()->canvasMap(c->yAxis());
    const double px = xMap.invTransform(p.x());
    const double py = yMap.invTransform(p.y());
    QUndoCommand *cmd = new AddCommand(this, c, index, px, py); //MemLeak?
    undoStack->push(cmd);
}

// Delete the selected point
void QwtPlotEdit::deletePoint()
{
    if (!d_select.isPoint())
        return;
    QwtPlotCurve *c = d_select.c;
    const int dataSize = c->dataSize();
    if (dataSize < 1) return;
    int index = d_select.index;
    // disable removal of first and last points
    if (d_select.index == 0 || d_select.index == dataSize-1) return;
    if (d_select.index > 1)
        d_select.index--;
    QUndoCommand *cmd = new RemoveCommand(this, c, index); //MemLeak?
    undoStack->push(cmd);
}

// Move the selected point
void QwtPlotEdit::moveBy(int dx, int dy)
{
    if ( dx == 0 && dy == 0 )
        return;

    if ( d_select.empty() )
        return;

    const int x = plot()->transform(d_select.c->xAxis(),
                                    d_select.c->x(d_select.index)) + dx;
    const int y = plot()->transform(d_select.c->yAxis(),
                                    d_select.c->y(d_select.index)) + dy;

    move(QPoint(x, y));
}

// Move the selected point
void QwtPlotEdit::move(const QPoint &pos)
{
    if ( d_select.empty() )
        return;
    int index = d_select.index;
    double curveX = d_select.c->x(index);
    double curveY = d_select.c->y(index);
    const double dx = plot()->invTransform(d_select.c->xAxis(), pos.x()) - curveX;
    const double dy = plot()->invTransform(d_select.c->yAxis(), pos.y()) - curveY;
    QUndoCommand *cmd = new MoveCommand(this, d_select.c, d_select.index, dx, dy); //MemLeak?
    undoStack->push(cmd);
}

void QwtPlotEdit::updateSelection()
{
    //    showCursor(false);
    //    plot()->replot();
    showCursor(true);
}

// Hightlight the selected point
void QwtPlotEdit::newSelection(CurveSelection select)
{
    showCursor(false);
    d_select = select;
    showCursor(true);
}

void QwtPlotEdit::showCursor(bool showIt)
{
    if (d_select.empty()) return;
    const QwtSymbol symbol = d_select.c->symbol();

    QwtSymbol newSymbol = symbol;
    if ( showIt ) {
        newSymbol.setBrush(symbol.brush().color().dark(150));
        if (d_select.isSegment())
            newSymbol.setBrush(symbol.brush().color().light(150));
    }

    const bool doReplot = plot()->autoReplot();

    plot()->setAutoReplot(false);
    d_select.c->setSymbol(newSymbol);

    if (d_select.isPoint())
        d_select.c->draw(d_select.index, d_select.index);
    else
        if (d_select.isSegment())
            d_select.c->draw(d_select.index, d_select.index+1);

    d_select.c->setSymbol(symbol);
    plot()->setAutoReplot(doReplot);
}

// Select the next/previous curve
void QwtPlotEdit::shiftCurveCursor(bool up)
{
    QwtPlotItemIterator it;

    const QwtPlotItemList &itemList = plot()->itemList();

    QwtPlotItemList curveList;
    for ( it = itemList.begin(); it != itemList.end(); ++it )
    {
        if ( (*it)->rtti() == QwtPlotItem::Rtti_PlotCurve )
            curveList += *it;
    }
    if ( curveList.isEmpty() )
        return;

    it = curveList.begin();

    if ( !d_select.empty() )
    {
        for ( it = curveList.begin(); it != curveList.end(); ++it )
        {
            if ( d_select.c == *it )
                break;
        }
        if ( it == curveList.end() ) // not found
            it = curveList.begin();

        if ( up )
        {
            ++it;
            if ( it == curveList.end() )
                it = curveList.begin();
        }
        else
        {
            if ( it == curveList.begin() )
                it = curveList.end();
            --it;
        }
    }

    showCursor(false);
    d_select.index = 0;
    d_select.c = (QwtPlotCurve *)*it;
    showCursor(true);
}

// Select the next/previous neighbour of the selected point
void QwtPlotEdit::shiftPointCursor(bool up)
{
    if ( d_select.empty() )
        return;

    int index = d_select.index + (up ? 1 : -1);
    index = (index + d_select.c->dataSize()) % d_select.c->dataSize();

    if ( index != d_select.index )
    {
        showCursor(false);
        d_select.index = index;
        showCursor(true);
    }
}
